"""Algorithm support for grid to polygon conversion.

See Also
--------
:meth:`pycontrails.MetDataArray.to_polygon_feature`
:meth:`pycontrails.MetDataArray.to_polygon_feature_collection`
"""

from __future__ import annotations

import warnings
from typing import Any

import numpy as np
import numpy.typing as npt

try:
    import cv2
    import shapely
    import shapely.geometry
    import shapely.validation
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "This module requires the 'opencv-python' and 'shapely' packages. "
        "These can be installed with 'pip install pycontrails[vis]'."
    ) from exc


def buffer_and_clean(
    contour: npt.NDArray[np.float_],
    min_area: float,
    convex_hull: bool,
    epsilon: float,
    precision: int | None,
    buffer: float,
    is_exterior: bool,
) -> shapely.Polygon | None:
    """Buffer and clean a contour.

    Parameters
    ----------
    contour : npt.NDArray[np.float_]
        Contour to buffer and clean. A 2d array of shape (n, 2) where n is the number
        of vertices in the contour.
    min_area : float
        Minimum area of the polygon. If the area of the buffered contour is less than
        this, return None.
    convex_hull : bool
        Whether to take the convex hull of the buffered contour.
    epsilon : float
        Epsilon value for polygon simplification. If 0, no simplification is performed.
    precision : int | None
        Precision of the output polygon. If None, no rounding is performed.
    buffer : float
        Buffer distance.
    is_exterior : bool, optional
        Whether the contour is an exterior contour. If True, the contour is buffered
        with a larger buffer distance. The polygon orientation is CCW iff this is True.

    Returns
    -------
    shapely.Polygon | None
        Buffered and cleaned polygon. If the area of the buffered contour is less than
        ``min_area``, return None.
    """
    if len(contour) == 1:
        base = shapely.Point(contour)
    elif len(contour) < 4:
        base = shapely.LineString(contour)
    else:
        base = shapely.LinearRing(contour)

    if is_exterior:
        # The contours computed by openCV go directly over array points
        # with value 1. With marching squares, we expect the contours to
        # be the midpoint between the 0-1 boundary. Apply a small buffer
        # to the exterior contours to account for this.
        polygon = base.buffer(buffer, quad_segs=1)
    else:
        # Only buffer the interiors if necessary
        try:
            polygon = shapely.Polygon(base)
        except shapely.errors.TopologicalError:
            return None
        if not polygon.is_valid:
            polygon = polygon.buffer(buffer / 10, quad_segs=1)

    assert polygon.is_valid

    # Remove all interior rings
    if polygon.interiors:
        polygon = shapely.Polygon(polygon.exterior)

    if polygon.area < min_area:
        return None

    # Exterior polygons should have CCW orientation
    if is_exterior != polygon.exterior.is_ccw:
        polygon = polygon.reverse()
    if convex_hull:
        polygon = _take_convex_hull(polygon)
    if epsilon:
        polygon = _buffer_simplify_iterate(polygon, epsilon)

    if precision is not None:
        while precision < 10:
            out = _round_polygon(polygon, precision)
            if out.is_valid:
                return out
            precision += 1

        warnings.warn("Could not round polygon to a valid geometry.")

    return polygon


def _round_polygon(polygon: shapely.Polygon, precision: int) -> shapely.Polygon:
    """Round the coordinates of a polygon.

    Parameters
    ----------
    polygon : shapely.Polygon
        Polygon to round.
    precision : int
        Precision to use when rounding.

    Returns
    -------
    shapely.Polygon
        Polygon with rounded coordinates.
    """
    if polygon.is_empty:
        return polygon

    exterior = np.round(np.asarray(polygon.exterior.coords), precision)
    interiors = [np.round(np.asarray(i.coords), precision) for i in polygon.interiors]
    return shapely.Polygon(exterior, interiors)


def _contours_to_polygons(
    contours: tuple[npt.NDArray[np.float_], ...],
    hierarchy: npt.NDArray[np.int_],
    min_area: float,
    convex_hull: bool,
    epsilon: float,
    longitude: npt.NDArray[np.float_] | None,
    latitude: npt.NDArray[np.float_] | None,
    precision: int | None,
    buffer: float,
    i: int = 0,
) -> list[shapely.Polygon]:
    """Convert the outputs of :func:`cv2.findContours` to :class:`shapely.Polygon`.

    Parameters
    ----------
    contours : tuple[npt.NDArray[np.float_], ...]
        The contours output from :func:`cv2.findContours`.
    hierarchy : npt.NDArray[np.int_]
        The hierarchy output from :func:`cv2.findContours`.
    min_area : float
        Minimum area of a polygon to be included in the output.
    convex_hull : bool
        Whether to take the convex hull of each polygon.
    epsilon : float
        Epsilon value to use when simplifying the polygons.
    longitude : npt.NDArray[np.float_] | None
        Longitude values for the grid.
    latitude : npt.NDArray[np.float_] | None
        Latitude values for the grid.
    precision : int | None
        Precision to use when rounding the coordinates.
    buffer : float
        Buffer to apply to the contours when converting to polygons.
    i : int, optional
        The index of the contour to start with. Defaults to 0.

    Returns
    -------
    list[shapely.Polygon]
        A list of polygons. Polygons with a parent-child relationship are merged into
        a single polygon.
    """
    out = []
    while i != -1:
        child_i, parent_i = hierarchy[i, 2:]
        is_exterior = parent_i == -1

        contour = contours[i][:, 0, ::-1]
        i = hierarchy[i, 0]
        if longitude is not None and latitude is not None:
            lon_idx = contour[:, 0]
            lat_idx = contour[:, 1]

            # Calculate interpolated longitude and latitude values and recreate contour
            lon = np.interp(lon_idx, np.arange(longitude.shape[0]), longitude)
            lat = np.interp(lat_idx, np.arange(latitude.shape[0]), latitude)
            contour = np.stack([lon, lat], axis=1)

        polygon = buffer_and_clean(
            contour,
            min_area,
            convex_hull,
            epsilon,
            precision,
            buffer,
            is_exterior,
        )
        if polygon is None:
            continue

        if child_i != -1:
            holes = _contours_to_polygons(
                contours,
                hierarchy,
                min_area=min_area,
                convex_hull=False,
                epsilon=epsilon,
                longitude=longitude,
                latitude=latitude,
                precision=precision,
                buffer=buffer,
                i=child_i,
            )

            candidate = shapely.Polygon(polygon.exterior, [h.exterior for h in holes])
            # Abundance of caution: check if the candidate is valid
            # If the candidate isn't valid, ignore all the holes
            # This can happen if there are many holes and the buffer operation
            # causes the holes to overlap
            if candidate.is_valid:
                polygon = candidate

        out.append(polygon)
    return out


def determine_buffer(longitude: npt.NDArray[np.float_], latitude: npt.NDArray[np.float_]) -> float:
    """Determine the proper buffer size to use when converting to polygons."""
    try:
        d_lon = longitude[1] - longitude[0]
        d_lat = latitude[1] - latitude[0]
    except IndexError as e:
        raise ValueError("Longitude and latitude must each have at least 2 elements.") from e

    if d_lon != d_lat:
        warnings.warn(
            "Longitude and latitude are not evenly spaced. Buffer size may be inaccurate."
        )
    if not np.all(np.diff(longitude) == d_lon):
        warnings.warn("Longitude is not evenly spaced. Buffer size may be inaccurate.")
    if not np.all(np.diff(latitude) == d_lat):
        warnings.warn("Latitude is not evenly spaced. Buffer size may be inaccurate.")

    return min(d_lon, d_lat) / 2


def find_multipolygon(
    arr: npt.NDArray[np.float_],
    threshold: float,
    min_area: float,
    epsilon: float,
    interiors: bool = True,
    convex_hull: bool = False,
    longitude: npt.NDArray[np.float_] | None = None,
    latitude: npt.NDArray[np.float_] | None = None,
    precision: int | None = None,
) -> shapely.MultiPolygon:
    """Compute a multipolygon from a 2d array.

    Parameters
    ----------
    arr : npt.NDArray[np.float_]
        Array to convert to a multipolygon. The array will be converted to a binary
        array by comparing each element to `threshold`. This binary array is then
        passed into :func:`cv2.findContours` to find the contours.
    threshold : float
        Threshold to use when converting `arr` to a binary array.
    min_area : float
        Minimum area of a polygon to be included in the output.
    epsilon : float
        Epsilon value to use when simplifying the polygons. Passed into shapely's
        :meth:`shapely.geometry.Polygon.simplify` method.
    interiors : bool
        Whether to include interior polygons.
    convex_hull : bool
        Experimental. Whether to take the convex hull of each polygon.
    longitude, latitude : npt.NDArray[np.float_], optional
        If provided, the coordinates values corresponding to the dimensions of `arr`.
        The contour coordinates will be converted to longitude-latitude values by indexing
        into this array. Defaults to None.
    precision : int, optional
        If provided, the precision to use when rounding the coordinates. Defaults to None.

    Returns
    -------
    shapely.MultiPolygon
        A multipolygon of the contours.
    """
    if arr.ndim != 2:
        raise ValueError("Array must be 2d")
    assert (longitude is None) == (latitude is None)
    if longitude is not None:
        assert latitude is not None
        assert arr.shape == (*longitude.shape, *latitude.shape)
        buffer = determine_buffer(longitude, latitude)
    else:
        buffer = 0.5

    arr_bin = np.empty(arr.shape, dtype=np.uint8)
    np.greater_equal(arr, threshold, out=arr_bin)

    mode = cv2.RETR_CCOMP if interiors else cv2.RETR_EXTERNAL
    contours, hierarchy = cv2.findContours(arr_bin, mode, cv2.CHAIN_APPROX_SIMPLE)
    if not contours:
        return shapely.MultiPolygon()

    assert len(hierarchy) == 1
    hierarchy = hierarchy[0]

    polygons = _contours_to_polygons(
        contours,
        hierarchy,
        min_area,
        convex_hull,
        epsilon,
        longitude,
        latitude,
        precision,
        buffer,
    )
    return _make_valid_multipolygon(polygons, convex_hull)


def _take_convex_hull(polygon: shapely.Polygon) -> shapely.Polygon:
    """Take the convex hull of a linear ring and preserve the orientation.

    Parameters
    ----------
    polygon : shapely.Polygon
        Linear ring to take the convex hull of.

    Returns
    -------
    shapely.Polygon
        Convex hull of the input.
    """
    convex_hull = polygon.convex_hull
    if polygon.exterior.is_ccw == convex_hull.exterior.is_ccw:
        return convex_hull
    return shapely.Polygon(convex_hull.exterior.reverse())


def _buffer_simplify_iterate(polygon: shapely.Polygon, epsilon: float) -> shapely.Polygon:
    """Simplify a linear ring by iterating over a larger buffer.

    This function calls :func:`shapely.buffer` and :func:`shapely.simplify`
    over a range of buffer distances. The buffer allows for the topology
    of the contour to change, which is useful for simplifying contours.
    Applying a buffer does introduce a slight bias towards the exterior
    of the contour.

    .. versionadded:: 0.38.0

    Parameters
    ----------
    polygon : shapely.Polygon
        Linear ring to simplify.
    epsilon : float
        Passed as ``tolerance`` parameter into :func:`shapely.simplify`.

    Returns
    -------
    shapely.LinearRing
        Simplified linear ring.
    """
    # Try to simplify without a buffer first
    # This seems to be computationally faster
    out = polygon.simplify(epsilon, preserve_topology=False)

    # In some rare situations, calling simplify actually gives a MultiPolygon.
    # In this case, take the polygon with the largest area
    if isinstance(out, shapely.MultiPolygon):
        out = max(out.geoms, key=lambda x: x.area)

    if out.is_simple and out.is_valid:
        return out

    # Applying a naive linear_ring.buffer(0) can destroy the polygon completely
    # https://stackoverflow.com/a/20873812

    is_ccw = polygon.exterior.is_ccw

    # Values here are somewhat ad hoc: These seem to allow the algorithm to
    # terminate and are not too computationally expensive
    for i in range(1, 11):
        distance = epsilon * i / 10

        # Taking the buffer can change the orientation of the contour
        out = polygon.buffer(distance, quad_segs=1)
        if out.exterior.is_ccw != is_ccw:
            out = shapely.Polygon(out.exterior.coords[::-1])

        out = out.simplify(epsilon, preserve_topology=False)
        if out.is_simple and out.is_valid:
            return out

    warnings.warn(
        f"Could not simplify contour with epsilon {epsilon}. Try passing a smaller epsilon."
    )
    return polygon.simplify(epsilon, preserve_topology=True)


def _make_valid_multipolygon(
    polygons: list[shapely.Polygon], convex_hull: bool
) -> shapely.MultiPolygon:
    """Make a multipolygon valid.

    This function attempts to make a multipolygon valid by iteratively
    applying :func:`shapely.unary_union` to convert non-disjoint polygons
    into disjoint polygons by merging them. If the multipolygon is still
    invalid after 5 attempts, a warning is raised and the last
    multipolygon is returned.

    This function is needed because simplifying a contour can change the
    geometry of it, which can cause non-disjoint polygons to be created.

    .. versionadded:: 0.38.0

    Parameters
    ----------
    polygons : list[shapely.Polygon]
        List of polygons to combine into a multipolygon.
    convex_hull : bool
        If True, take the convex hull of merged polygons.

    Returns
    -------
    shapely.MultiPolygon
        Valid multipolygon.
    """
    mp = shapely.MultiPolygon(polygons)
    if mp.is_empty:
        return mp

    n_attempts = 5
    for _ in range(n_attempts):
        if mp.is_valid:
            return mp

        mp = shapely.unary_union(mp)
        if isinstance(mp, shapely.Polygon):
            mp = shapely.MultiPolygon([mp])

        # Fix the orientation of the polygons
        mp = shapely.MultiPolygon([shapely.geometry.polygon.orient(p, sign=1) for p in mp.geoms])

        if convex_hull:
            mp = shapely.MultiPolygon([_take_convex_hull(p) for p in mp.geoms])

    warnings.warn(
        f"Could not make multipolygon valid after {n_attempts} attempts. "
        "According to shapely, the multipolygon is invalid because: "
        f"{shapely.validation.explain_validity(mp)}"
    )
    return mp


def multipolygon_to_geojson(
    multipolygon: shapely.MultiPolygon,
    altitude: float | None,
    properties: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Convert a shapely multipolygon to a GeoJSON feature.

    Parameters
    ----------
    multipolygon : shapely.MultiPolygon
        Multipolygon to convert.
    altitude : float or None
        Altitude of the multipolygon. If provided, the multipolygon coordinates
        will be given a z-coordinate.
    properties : dict[str, Any], optional
        Properties to add to the GeoJSON feature.

    Returns
    -------
    dict[str, Any]
        GeoJSON feature with geometry type "MultiPolygon".
    """
    coordinates = []
    for polygon in multipolygon.geoms:
        poly_coords = []
        rings = polygon.exterior, *polygon.interiors
        for ring in rings:
            if altitude is None:
                coords = np.asarray(ring.coords)
            else:
                shape = len(ring.coords), 3
                coords = np.empty(shape)
                coords[:, :2] = ring.coords
                coords[:, 2] = altitude

            poly_coords.append(coords.tolist())
        coordinates.append(poly_coords)

    return {
        "type": "Feature",
        "properties": properties or {},
        "geometry": {"type": "MultiPolygon", "coordinates": coordinates},
    }



"""GFS unit tests."""

from __future__ import annotations

from datetime import datetime

import pytest

from pycontrails import MetVariable
from pycontrails.core.met_var import AirTemperature
from pycontrails.datalib.gfs import GFS_VARIABLES, GFSForecast


def test_GFS_parameters() -> None:
    """Test GFS."""
    assert all([isinstance(p, MetVariable) for p in GFS_VARIABLES])


def test_GFSForecast_init() -> None:
    """Test GFSForecast init."""

    times = ("2022-03-01T03:00:00", "2022-03-01T06:30:00")
    gfs = GFSForecast(times, variables=["air_temperature"], pressure_levels=[300, 250])
    assert gfs.forecast_time == datetime(2022, 3, 1, 0)
    assert gfs.step_offset == 3
    assert gfs.steps == [3, 4, 5, 6, 7]
    assert gfs.variables == [AirTemperature]
    assert gfs.grid == 0.25

    # forecast time jumps every 6 hours
    times = ("2022-03-01T08:00:00", "2022-03-01T010:30:00")
    gfs = GFSForecast(times, variables=["air_temperature"], pressure_levels=[300, 250])
    assert gfs.forecast_time == datetime(2022, 3, 1, 6)

    times = ("2022-03-01T13:00:00", "2022-03-01T014:30:00")
    gfs = GFSForecast(times, variables=["air_temperature"], pressure_levels=[300, 250])
    assert gfs.forecast_time == datetime(2022, 3, 1, 12)

    times = ("2022-03-01T19:00:00", "2022-03-01T020:30:00")
    gfs = GFSForecast(times, variables=["air_temperature"], pressure_levels=[300, 250])
    assert gfs.forecast_time == datetime(2022, 3, 1, 18)

    # unsupported pressure levels
    with pytest.raises(ValueError):
        GFSForecast(time=[], variables=[], pressure_levels=[17, 18, 19])

    # grid strings
    times = ("2022-03-01T03:00:00", "2022-03-01T06:30:00")
    gfs = GFSForecast(times, variables=["air_temperature"], pressure_levels=[300, 250], grid=0.5)
    assert gfs.forecast_time == datetime(2022, 3, 1, 0)
    assert gfs.grid == 0.5
    assert gfs.step_offset == 3
    assert gfs.steps == [3, 4, 5, 6, 7]
    assert gfs.variables == [AirTemperature]
    assert gfs.grid == 0.5


def test_GFSForecast_repr() -> None:
    """Test GFSForecast __repr__."""
    gfs = GFSForecast(
        time=datetime(2000, 1, 1), variables="ice_water_mixing_ratio", pressure_levels=200
    )
    out = repr(gfs)
    assert "GFS" in out
    assert "Forecast time:" in out


def test_GFSForecast_grid_string() -> None:
    """Test GFSForecast _grid_string property."""

    # grid strings
    times = ("2022-03-01T03:00:00", "2022-03-01T06:30:00")
    gfs = GFSForecast(times, variables=["air_temperature"], pressure_levels=[300, 250], grid=0.5)
    assert gfs.grid == 0.5
    assert gfs._grid_string == "0p50"

    gfs = GFSForecast(times, variables=["air_temperature"], pressure_levels=[300, 250], grid=1.0)
    assert gfs.grid == 1
    assert gfs._grid_string == "1p00"


def test_GFSForecast_steps() -> None:
    """Test GFSForecast steps property."""

    times = ("2022-03-01T00:00:00", "2022-03-01T03:00:00")
    gfs = GFSForecast(times, variables="t", pressure_levels=[300, 250])
    assert gfs.steps == [0, 1, 2, 3]

    times = ("2022-03-01T02:00:00", "2022-03-01T07:00:00")
    gfs = GFSForecast(times, variables="t", pressure_levels=[300, 250])
    assert gfs.steps == [2, 3, 4, 5, 6, 7]

    # from the 06 forecast
    times = ("2022-03-01T06:00:00", "2022-03-01T08:00:00")
    gfs = GFSForecast(times, variables="t", pressure_levels=[300, 250])
    assert gfs.steps == [0, 1, 2]


def test_GFSForecast_forecast_path() -> None:
    """Test GFSForecast forecast path construction."""

    gfs = GFSForecast(time="2022-03-01T014:00:00", variables="t", pressure_levels=200)
    assert gfs.forecast_path == "gfs.20220301/12/atmos"


def test_GFSForecast_filenames() -> None:
    """Test GFSForecast filename construction."""

    times = ("2022-03-01T06:00:00", "2022-03-01T08:00:00")
    gfs = GFSForecast(time=times, variables="t", pressure_levels=200)

    # get filenames
    filename = gfs.filename(gfs.steps[0])
    assert filename == "gfs.t06z.pgrb2.0p25.f000"

    filename = gfs.filename(gfs.steps[1])
    assert filename == "gfs.t06z.pgrb2.0p25.f001"

    filename = gfs.filename(gfs.steps[2])
    assert filename == "gfs.t06z.pgrb2.0p25.f002"


def test_GFSForecast_hash() -> None:
    """Test GFSForecast hash string."""

    times = ("2022-03-01T06:00:00", "2022-03-01T08:00:00")
    gfs = GFSForecast(time=times, variables="t", pressure_levels=200)
    gfs2 = GFSForecast(time=times, variables=["t", "r"], pressure_levels=200)
    gfs3 = GFSForecast(time=times, variables=["t", "r"], pressure_levels=200, grid=0.5)

    assert gfs.hash == "2e7699eb113d903633823b858058c37a74264a98"
    assert gfs.hash != gfs2.hash
    assert gfs2.hash != gfs3.hash


def test_GFSForecast_cachepaths() -> None:
    """Test GFSForecast cachepath construction."""

    times = ("2022-03-01T06:00:00", "2022-03-01T08:00:00")
    gfs = GFSForecast(time=times, variables="t", pressure_levels=200)

    # get cachepaths for all timestaps
    cachepaths = [gfs.create_cachepath(t) for t in gfs.timesteps]

    assert "20220301-06-0-gfspl0.25.nc" in cachepaths[0]
    assert "20220301-06-1-gfspl0.25.nc" in cachepaths[1]



"""ECMWF Data Access."""

from __future__ import annotations

from pycontrails.datalib.spire.spire import (
    clean,
    generate_flight_id,
    identify_flights,
    is_valid_trajectory,
    validate_flights,
)

__all__ = [
    "clean",
    "generate_flight_id",
    "identify_flights",
    "is_valid_trajectory",
    "validate_flights",
]



"""Test geo module."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pyproj
import pytest

from pycontrails.physics import constants, geo

from .conftest import get_static_path


@pytest.fixture
def rand_geo_data():
    rng = np.random.default_rng(1234)
    n_pts = 100000
    lons = -180 + 360 * rng.random(n_pts)
    lats = -90 + 180 * rng.random(n_pts)
    az = 360 * rng.random(n_pts)
    dist = 100000 * (rng.random(n_pts) - 0.5)  # can be negative
    return {"lons": lons, "lats": lats, "az": az, "dist": dist}


@pytest.fixture
def geod() -> pyproj.Geod:
    return pyproj.Geod(a=constants.radius_earth)


def test_haversine(geod: pyproj.Geod, rand_geo_data: dict[str, np.ndarray]):
    """Check that `geo.segment_haversine` agrees with `pyproj` implementation."""
    lons = rand_geo_data["lons"]
    lats = rand_geo_data["lats"]

    pyproj_lengths = geod.line_lengths(lons, lats)

    # our version of haversine tacks an extra nan onto the end
    pycont_lengths = geo.segment_haversine(lons, lats)
    assert np.isnan(pycont_lengths[-1])
    np.testing.assert_allclose(pyproj_lengths, pycont_lengths[:-1])

    # test that haversine wraps longitude
    d_meridian = geo.haversine(-180, 5, 180, 5)
    assert np.round(d_meridian, 6) == 0

    d_meridian = geo.haversine(-180, 5, 179.9, 5)
    assert np.round(d_meridian, 6) == 11077.577787

    d_meridian = geo.haversine(-179.9, 5, 180, 5)
    assert np.round(d_meridian, 6) == 11077.577787


def test_forward_azimuth(geod: pyproj.Geod, rand_geo_data: dict[str, np.ndarray]):
    """Check that `geo.forward_azimuth` agrees with `pyproj` implementation."""
    # pyproj also returns "back azimuth", which is ignored
    pyproj_lons, pyproj_lats, _ = geod.fwd(**rand_geo_data)

    pycont_lons, pycont_lats = geo.forward_azimuth(**rand_geo_data)

    np.testing.assert_allclose(pyproj_lons, pycont_lons)
    np.testing.assert_allclose(pyproj_lats, pycont_lats)


def test_azimuth_to_direction(geod: pyproj.Geod, rand_geo_data: dict[str, np.ndarray]):
    """Check that `geo.azimuth_to_direction` agrees with `pyproj` implementation."""
    longitude = rand_geo_data["lons"]
    latitude = rand_geo_data["lats"]
    azimuth = rand_geo_data["az"]
    pycont_sin_a, pycont_cos_a = geo.azimuth_to_direction(azimuth, latitude)

    dist = np.ones_like(longitude)
    fwd_lon, fwd_lat, _ = geod.fwd(longitude, latitude, azimuth, dist)
    adj = fwd_lon - longitude
    opp = fwd_lat - latitude
    hyp = np.sqrt(adj * adj + opp * opp)
    pyproj_sin_a = opp / hyp
    pyproj_cos_a = adj / hyp

    # If we use a smaller dist above, the error here would be smaller
    np.testing.assert_allclose(pycont_sin_a, pyproj_sin_a, atol=1e-4)
    np.testing.assert_allclose(pycont_cos_a, pyproj_cos_a, atol=1e-4)


def test_segment_angle(geod: pyproj.Geod):
    rng = np.random.default_rng(123)  # seed chosen to get decent coverage in random walk
    longitude = 0.4 * rng.random(100000) - 0.2
    latitude = 0.4 * rng.random(100000) - 0.2

    # longitude-latitude random walk
    longitude = np.cumsum(longitude)
    latitude = np.cumsum(latitude)

    # stays within bounds
    assert latitude.max() < 90 and latitude.min() > -90

    sin, cos = geo.segment_angle(longitude, latitude)
    assert np.isnan(cos[-1])
    assert np.isnan(sin[-1])

    angle, _, _ = geod.inv(
        lons1=longitude[:-1],
        lats1=latitude[:-1],
        lons2=longitude[1:],
        lats2=latitude[1:],
    )
    angle = np.deg2rad(90 - angle)
    sin2 = np.sin(angle)
    cos2 = np.cos(angle)

    np.testing.assert_array_almost_equal(sin[:-1], sin2, decimal=10)
    np.testing.assert_array_almost_equal(cos[:-1], cos2, decimal=10)

    sin3, cos3 = deprecated_segment_angle(longitude, latitude)

    # Depreciated implemention is not mathematically, but still agrees to
    # 2 decimal places of accuracy. Depreciated version does not use the
    # append nan convention
    np.testing.assert_array_almost_equal(sin[:-1], sin3[:-1], decimal=2)
    np.testing.assert_array_almost_equal(cos[:-1], cos3[:-1], decimal=2)


def test_segment_angle_deprecated_diverge():
    # deprecated version differs from current implementation for large segments

    coords = np.array([20, 50])
    sin, cos = geo.segment_angle(coords, coords)
    sin2, cos2 = deprecated_segment_angle(coords, coords)

    # final coordinate (convention) is different
    assert np.isnan(cos[1])
    assert np.isnan(sin[1])
    assert sin2[1] == 0
    assert cos2[1] == 1

    # first coordinate differs
    with pytest.raises(AssertionError):
        np.testing.assert_array_almost_equal(sin, sin2, decimal=2)

    with pytest.raises(AssertionError):
        np.testing.assert_array_almost_equal(cos, cos2, decimal=2)


def deprecated_segment_angle(
    longitude: np.ndarray, latitude: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    """
    OLD VERSION OF `geo.segment_angle` PORTED FROM FORTRAN
    THIS FUNCTION NEARLY AGREES WITH CURRENT IMPLEMENTATION FOR SMALL SEGMENTS
    CURRENT IMPLEMENTATION IS MATHEMATICALLY CORRECT AND TWICE AS FAST

    Calculates the sin(a) and cos(a) for the
    angle between each segment and the longitudinal axis


                (lon_2, lat_2)  X
                               /|
                              / |
                             /  |
                            /   |
                           /    |
                          /     |
                         /      |
        (lon_1, lat_1)  X -------> longitude (x-axis)

    Returns
    -------
    np.ndarray, np.ndarray
        sin(a), cos(a), where a is the angle between the segment and the longitudinal axis

    See Also
    --------
    :func:`haversine`
    """
    from pycontrails.physics import constants, units

    lats_next = np.roll(latitude, -1)
    lats_avg = 0.5 * (latitude + lats_next)

    # append nan to ensure length is the same as the original waypoints
    d_lon = np.append(np.diff(longitude), np.nan)
    d_lat = np.append(np.diff(latitude), np.nan)

    dist = geo.segment_haversine(longitude, latitude)
    sin_a = (constants.radius_earth * units.degrees_to_radians(d_lat)) / dist
    sin_a[-1] = 0  # last element is a discontinuous segment
    assert isinstance(sin_a, np.ndarray)

    cos_a = (
        constants.radius_earth
        * units.degrees_to_radians(d_lon)
        * np.cos(units.degrees_to_radians(lats_avg))
        / dist
    )
    cos_a[-1] = 1  # last element is a discontinuous segment

    return sin_a, cos_a


def test_latitude_advection_divergence():
    """Show that it is possible for latitude values to diverge near poles.

    This test highlights polar divergence arising in a `CocipGrid` simulation.
    """
    rng = np.random.default_rng(2468)
    latitude = rng.uniform(89.3, 89.7, 1000)  # start with latitude near 89.5
    v_wind = rng.uniform(-20, 20, 1000)  # with typical wind values
    dt = np.timedelta64(30, "m")
    latitude_t2 = geo.advect_latitude(latitude=latitude, v_wind=v_wind, dt=dt)
    assert np.any(latitude_t2 > 90)  # conclude some latitude values exceed 90


def test_solar_calculations() -> None:
    """Test solar calculations against NOAA Solar Calculator.

    See NOAA_Solar_Calculations_day.xls from https://gml.noaa.gov/grad/solcalc/calcdetails.html.
    """

    noaa_solar_csv = pd.read_csv(get_static_path("NOAA_Solar_Calculations_day.csv"))

    time = (
        pd.to_datetime(noaa_solar_csv["Date"])
        + pd.to_timedelta(noaa_solar_csv["Time (past local midnight)"])
    ).to_numpy()
    latitude = np.full(time.shape, fill_value=40.0)
    longitude = np.full(time.shape, fill_value=-105.0)

    # orbital position
    theta_rad = geo.orbital_position(time)
    # TODO: comparison column?

    # orbital correction
    # orbital_correction = geo.orbital_correction_for_solar_hour_angle(theta_rad)
    # TODO: comparison column?

    # solar zenith angle
    noaa_sza = noaa_solar_csv["Solar Zenith Angle (deg)"]
    cos_sza = geo.cosine_solar_zenith_angle(longitude, latitude, time, theta_rad)
    sza = np.arccos(cos_sza) * 180 / np.pi
    np.testing.assert_allclose(sza, noaa_sza, atol=1e-1)

    # solar declination
    noaa_sda = noaa_solar_csv["Sun Declin (deg)"]
    sda = geo.solar_declination_angle(theta_rad)
    np.testing.assert_allclose(sda, noaa_sda, atol=1e-1)

    # solar hour angle
    # TODO: the value diverge near midnight
    noaa_sha = noaa_solar_csv["Hour Angle (deg)"]
    sha = geo.solar_hour_angle(longitude, time, theta_rad)
    sha[sha < -180] = sha[sha < -180] + 360
    np.testing.assert_allclose(sha, noaa_sha, atol=3e-1)


def test_azimuth(geod: pyproj.Geod, rand_geo_data: dict[str, np.ndarray]):
    """Check that `geo.azimuth` agrees with `pyproj` implementation."""

    longitude = rand_geo_data["lons"]
    latitude = rand_geo_data["lats"]

    azimuth = geo.segment_azimuth(longitude, latitude)
    assert np.isnan(azimuth[-1])

    angle, _, _ = geod.inv(
        lons1=longitude[:-1],
        lats1=latitude[:-1],
        lons2=longitude[1:],
        lats2=latitude[1:],
    )

    # reset output [0, 360)
    angle = angle % 360
    np.testing.assert_array_almost_equal(azimuth[:-1], angle, decimal=10)

    # this should be 90 degrees off from the logitudinal angle
    # so sin(azimuth) == cos(a), cos(azimuth) == sin(a)
    sin_azimuth = np.sin(np.deg2rad(azimuth))
    cos_azimuth = np.cos(np.deg2rad(azimuth))
    sin_a, cos_a = geo.segment_angle(longitude, latitude)

    np.testing.assert_array_almost_equal(sin_azimuth, cos_a, decimal=10)
    np.testing.assert_array_almost_equal(cos_azimuth, sin_a, decimal=10)



"""`Spire Aviation <https://spire.com/aviation/>`_ ADS-B data support."""

from __future__ import annotations

import logging

import numpy as np
import pandas as pd

from pycontrails.core import flight
from pycontrails.physics import geo, units

logger = logging.getLogger(__name__)

#: Minimum messages to identify a single flight trajectory
TRAJECTORY_MINIMUM_MESSAGES = 10

#: Data types of parsed message fields
#: "timestamp" is excluded as its parsed by :func:`pandas.to_datetime`
MESSAGE_DTYPES = {
    "icao_address": str,
    "latitude": float,
    "longitude": float,
    "altitude_baro": float,
    "heading": float,
    "speed": float,
    "on_ground": bool,
    "callsign": str,
    "tail_number": str,
    "collection_type": str,
    "aircraft_type_icao": str,
    "aircraft_type_name": str,
    "airline_iata": str,
    "airline_name": str,
    "departure_utc_offset": str,
    "departure_scheduled_time": str,
}


def clean(messages: pd.DataFrame) -> pd.DataFrame:
    """
    Remove erroneous messages from raw Spire ADS-B data.

    Copies input `messages` before modifying.

    Parameters
    ----------
    messages : pd.DataFrame
        Raw ADS-B messages

    Returns
    -------
    pd.DataFrame
        ADS-B messages with erroneous data removed.

    Notes
    -----
    This function removes:

    #. Remove messages without data in "icao_address", "aircraft_type_icao",
       "altitude_baro", "on_ground", "tail_number"
    #. Ensure columns have the dtype as specified in :attr:`MESSAGE_DTYPES`
    #. Remove messages with tail number "VARIOUS"
       and aircraft types "N/A", "GNDT", "GRND", "ZZZZ"
    #. Remove terrestrial messages without callsign.
       Most of these messages are below 10,000 feet and from general aviation.
    #. Remove messages when "on_ground" indicator is True, but
       speed > :attr:`flight.MAX_ON_GROUND_SPEED` knots
       or altitude > :attr:`flight.MAX_AIRPORT_ELEVATION` ft
    #. Drop duplicates by "icao_address" and "timestamp"
    """
    _n_messages = len(messages)

    # TODO: Enable method to work without copying
    # data multiple times
    mdf = messages.copy()

    # Remove waypoints without data in "icao_address", "aircraft_type_icao",
    # "altitude_baro", "on_ground", "tail_number"
    non_null_cols = [
        "timestamp",
        "longitude",
        "latitude",
        "altitude_baro",
        "on_ground",
        "icao_address",
        "aircraft_type_icao",
        "tail_number",
    ]
    mdf = mdf.dropna(subset=non_null_cols)

    # Ensure columns have the correct dtype
    mdf = mdf.astype(MESSAGE_DTYPES)

    # convert timestamp into a timezone naive pd.Timestamp
    mdf["timestamp"] = pd.to_datetime(mdf["timestamp"]).dt.tz_localize(None)

    # Remove messages with tail number set to VARIOUS
    # or aircraft type set to "N/A", "GNDT", "GRND", "ZZZZ"
    filt = (mdf["tail_number"] == "VARIOUS") | (
        mdf["aircraft_type_icao"].isin(["N/A", "GNDT", "GRND", "ZZZZ"])
    )
    mdf = mdf[~filt]

    # Remove terrestrial waypoints without callsign
    # Most of these waypoints are below 10,000 feet and from general aviation
    filt = (mdf["callsign"].isna()) & (mdf["collection_type"] == "terrestrial")
    mdf = mdf[~filt]

    # Fill missing callsigns for satellite records
    callsigns_missing = (mdf["collection_type"] == "satellite") & (mdf["callsign"].isna())
    mdf.loc[callsigns_missing, "callsign"] = None  # reset values to be None

    callsigns_missing_unique = mdf.loc[callsigns_missing, "icao_address"].unique()

    rows_with_any_missing_callsigns = mdf.loc[
        mdf["icao_address"].isin(callsigns_missing_unique),
        ["icao_address", "callsign", "collection_type"],
    ]

    for _, gp in rows_with_any_missing_callsigns.groupby("icao_address", sort=False):
        mdf.loc[gp.index, "callsign"] = gp["callsign"].fillna(method="ffill").fillna(method="bfill")

    # Remove messages with erroneous "on_ground" indicator
    filt = mdf["on_ground"] & (
        (mdf["speed"] > flight.MAX_ON_GROUND_SPEED)
        | (mdf["altitude_baro"] > flight.MAX_AIRPORT_ELEVATION)
    )
    mdf = mdf[~filt]

    # Drop duplicates by icao_address and timestamp
    mdf = mdf.drop_duplicates(subset=["icao_address", "timestamp"])

    logger.debug(f"{len(mdf) / _n_messages:.2f} messages remain after Spire ADS-B cleanup")

    return mdf.reset_index(drop=True)


def generate_flight_id(time: pd.Timestamp, callsign: str) -> str:
    """Generate a unique flight id for instance of flight.

    Parameters
    ----------
    time : pd.Timestamp
        First waypoint time associated with flight.
    callsign : str
        Callsign of the flight.
        Other flight identifiers could be used if the callsign
        is not defined.

    Returns
    -------
    str
        Flight id in the form "{%Y%m%d-%H%M}-{callsign}"
    """
    t_string = time.strftime("%Y%m%d-%H%M")
    return f"{t_string}-{callsign}"


def identify_flights(messages: pd.DataFrame) -> pd.Series:
    """Identify unique flights from Spire ADS-B messages.

    Parameters
    ----------
    messages : pd.DataFrame
        Cleaned ADS-B messages,
        as output from :func:`clean`

    Returns
    -------
    pd.Series
        Flight ids for the same index as `messages`

    Notes
    -----
    The algorithm groups flights initially on "icao_address".

    For each group:

    #. Fill callsign for satellite messages
    #. Group again by "tail_number", "aircraft_type_icao", "callsign"
    #. Remove flights with less than :attr:`TRAJECTORY_MINIMUM_MESSAGES` messages
    #. Separate flights by "on_ground" indicator. See `_separate_by_on_ground`.
    #. Separate flights by cruise phase. See `_separate_by_cruise_phase`.

    See Also
    --------
    :func:`clean`
    :class:`Spire`
    :func:`_separate_on_ground`
    :func:`_separate_by_cruise_phase`
    """

    # Set default flight id
    flight_id = pd.Series(
        data=None,
        dtype=object,
        index=messages.index,
    )

    for idx, gp in messages[
        [
            "icao_address",
            "tail_number",
            "aircraft_type_icao",
            "callsign",
            "timestamp",
            "longitude",
            "latitude",
            "altitude_baro",
            "on_ground",
        ]
    ].groupby(["icao_address", "tail_number", "aircraft_type_icao", "callsign"], sort=False):
        # minimum # of messages > TRAJECTORY_MINIMUM_MESSAGES
        if len(gp) < TRAJECTORY_MINIMUM_MESSAGES:
            logger.debug(f"Message {idx} group too small to create flight ids")
            continue

        # TODO: this altitude cleanup does not persist back into messages
        # this should get moved into flight module
        gp = _clean_trajectory_altitude(gp)

        # separate flights by "on_ground" column
        gp["flight_id"] = _separate_by_on_ground(gp)

        # further separate flights by cruise phase analysis
        for _, fl in gp.groupby("flight_id"):
            gp.loc[fl.index, "flight_id"] = _separate_by_cruise_phase(fl)

        # save flight ids
        flight_id.loc[gp.index] = gp["flight_id"]

    return flight_id


def _clean_trajectory_altitude(messages: pd.DataFrame) -> pd.DataFrame:
    """
    Clean erroneous and noisy altitude on a single flight.

    TODO: move this to Flight

    Parameters
    ----------
    messages: pd.DataFrame
        ADS-B messages from a single flight trajectory.

    Returns
    -------
    pd.DataFrame
        ADS-B messages with filtered altitude.

    Notes
    -----
    #. If ``pycontrails.ext.bada`` installed,
       remove erroneous altitude, i.e., altitude above operating limit of aircraft type
    #. Filter altitude signal to remove noise and
       snap cruise altitudes to 1000 ft intervals

    See Also
    --------
    :func:`flight.filter_altitude`
    """
    mdf = messages.copy()

    # Use BADA 3 to support filtering
    try:
        from pycontrails.ext.bada import BADA3

        bada3 = BADA3()
        aircraft_type_icao = (
            mdf["aircraft_type_icao"].iloc[0]
            if len(mdf["aircraft_type_icao"]) > 1
            else mdf["aircraft_type_icao"]
        )

        # Try remove aircraft types not covered by BADA 3 (mainly helicopters)
        mdf = mdf.loc[mdf["aircraft_type_icao"].isin(bada3.synonym_dict)]

        # Remove erroneous altitude, i.e., altitude above operating limit of aircraft type
        bada_max_altitude_ft = bada3.ptf_param_dict[aircraft_type_icao].max_altitude_ft
        is_above_ceiling = mdf["altitude_baro"] > bada_max_altitude_ft
        mdf = mdf.loc[~is_above_ceiling]

        # Set the min cruise altitude to 0.5 * "max_altitude_ft"
        min_cruise_altitude_ft = 0.5 * bada_max_altitude_ft

    except (ImportError, FileNotFoundError, KeyError):
        min_cruise_altitude_ft = flight.MIN_CRUISE_ALTITUDE

    # Filter altitude signal
    # See https://traffic-viz.github.io/api_reference/traffic.core.flight.html#traffic.core.Flight.filter
    mdf["altitude_baro"] = flight.filter_altitude(mdf["altitude_baro"].to_numpy())

    # Snap altitudes in cruise to the nearest flight level.
    # Requires segment phase
    altitude_ft = mdf["altitude_baro"].to_numpy()
    segment_duration = flight.segment_duration(mdf["timestamp"].to_numpy())
    segment_rocd = flight.segment_rocd(segment_duration, altitude_ft)
    segment_phase = flight.segment_phase(
        segment_rocd,
        altitude_ft,
        min_cruise_altitude_ft=min_cruise_altitude_ft,
    )
    is_cruise = segment_phase == flight.FlightPhase.CRUISE
    mdf.loc[is_cruise, "altitude_baro"] = np.round(altitude_ft[is_cruise], -3)

    return mdf


def _separate_by_on_ground(messages: pd.DataFrame) -> pd.Series:
    """Separate individual flights by "on_ground" column.

    The input ``messages`` are expected to grouped by "icao_address", "tail_number",
    "aircraft_type_icao", "callsign".

    Parameters
    ----------
    messages : pd.DataFrame
        Messages grouped by "icao_address", "tail_number",
        "aircraft_type_icao", "callsign".
        Must contain the "on_ground" and "altitude_baro" columns.

    Returns
    -------
    pd.Series
        Flight ids for the same index as `messages`
    """
    # Set default flight id
    try:
        flight_id = messages["flight_id"]
    except KeyError:
        flight_id = pd.Series(
            data=None,
            dtype=object,
            index=messages.index,
        )

    # make sure aircraft is actually on ground
    # TODO: use DEM for ground position?
    is_on_ground = messages["on_ground"] & (
        messages["altitude_baro"] < flight.MAX_AIRPORT_ELEVATION
    )

    # filter this signal so that it removes isolated 1-2 wide variations
    is_on_ground = is_on_ground.rolling(window=5).mean().bfill() > 0.5

    # find end of flight indexes using "on_ground"
    end_of_flight = (~is_on_ground).astype(int).diff(periods=-1) == -1

    # identify each individual flight using cumsum magic
    for _, gp in messages.groupby(end_of_flight.cumsum()):
        flight_id.loc[gp.index] = generate_flight_id(
            gp["timestamp"].iloc[0], gp["callsign"].iloc[0]
        )

    return flight_id


def _separate_by_cruise_phase(messages: pd.DataFrame) -> pd.Series:
    """
    Separate flights by multiple cruise phases.

    The input ``messages`` are expected to grouped by "icao_address", "tail_number",
    "aircraft_type_icao", "callsign".

    Its strongly encouraged to run :func:`flight.filter_altitude(...)` over the messages
    before passing into this function.

    Parameters
    ----------
    messages : pd.DataFrame
        Messages grouped by "icao_address", "tail_number",
        "aircraft_type_icao", "callsign".
        Must contain the "on_ground" and "altitude_baro" columns.

    Returns
    -------
    pd.Series
        Integer series for each unique flight with same index as "messages"

    Notes
    -----
    Flights with multiple cruise phases are identified by evaluating the
    messages between the start and end of the cruise phase.

    For these messages that should be at cruise, multiple cruise phases are identified when:

    #. Altitudes fall below the minimum cruise altitude.
       If pycontrails.ext.bada is installed, this value is set to half
       the BADA3 altitude ceiling of the aircraft.
       If not, this value is set to :attr:`flight.MIN_CRUISE_ALTITUDE`.
    #. There is a time difference > 15 minutes between messages.

    If multiple flights are identified,
    the cut-off point is specified at messages with the largest time difference.

    Flight "diversion" is defined when the aircraft descends below 10,000 feet
    and climbs back to cruise altitude to travel to the alternative airport.
    A diversion is identified when all five conditions below are satisfied:

    #. Altitude in any messages between
       the start and end of cruise is < :attr:`flight.MAX_AIRPORT_ELEVATION` ft
    #. Time difference between messages that should be
       at cruise must be < 15 minutes (continuous telemetry)
    #. Segment length between messages that should be
       at cruise must be > 500 m (no stationary messages),
    #. Time elapsed between message with the lowest altitude
       (during cruise) and final message should be < 2 h,
    #. No messages should be on the ground between the start and end of cruise.
    """

    # Set default flight id
    try:
        flight_id = messages["flight_id"].copy()
    except KeyError:
        flight_id = pd.Series(
            data=generate_flight_id(messages["timestamp"].iloc[0], messages["callsign"].iloc[0]),
            dtype=object,
            index=messages.index,
        )

    # Use BADA 3 to calculate max altitude
    try:
        from pycontrails.ext.bada import BADA3

        bada3 = BADA3()
        aircraft_type_icao = (
            messages["aircraft_type_icao"].iloc[0]
            if len(messages["aircraft_type_icao"]) > 1
            else messages["aircraft_type_icao"]
        )

        # Set the min cruise altitude to 0.5 * "max_altitude_ft"
        bada_max_altitude_ft = bada3.ptf_param_dict[aircraft_type_icao].max_altitude_ft
        min_cruise_altitude_ft = 0.5 * bada_max_altitude_ft

    except (ImportError, FileNotFoundError, KeyError):
        min_cruise_altitude_ft = flight.MIN_CRUISE_ALTITUDE

    # Calculate flight phase
    altitude_ft = messages["altitude_baro"].to_numpy()
    segment_duration = flight.segment_duration(messages["timestamp"].to_numpy())
    segment_rocd = flight.segment_rocd(segment_duration, altitude_ft)
    segment_phase = flight.segment_phase(
        segment_rocd,
        altitude_ft,
        threshold_rocd=250,
        min_cruise_altitude_ft=min_cruise_altitude_ft,
    )

    # get cruise phase
    cruise = segment_phase == flight.FlightPhase.CRUISE

    # fill between the first and last cruise phase indicator
    # this represents the flight phase after takeoff climb and before descent to landing
    # Index of first and final cruise waypoint
    # i_cruise_start = np.min(np.argwhere(flight_phase.cruise))
    # i_cruise_end = min(np.max(np.argwhere(flight_phase.cruise)), len(flight_phase.cruise))
    within_cruise = np.bitwise_xor.accumulate(cruise) | cruise

    # There should not be any waypoints with low altitudes between
    # the start and end of `within_cruise`
    is_low_altitude = altitude_ft < min_cruise_altitude_ft
    is_long_interval = segment_duration > (15 * 60)  # 15 minutes
    anomalous_phase = within_cruise & is_low_altitude & is_long_interval
    multiple_cruise_phase = np.any(anomalous_phase)

    # if there is only one cruise phase, just return one label
    if not multiple_cruise_phase:
        return flight_id

    # Check for presence of a diverted flight
    potentially_diverted = within_cruise & is_low_altitude
    if np.any(potentially_diverted):
        # Calculate segment length
        segment_length = geo.segment_length(
            messages["longitude"].to_numpy(),
            messages["latitude"].to_numpy(),
            units.ft_to_m(altitude_ft),
        )

        # get the index of the minimum altitude when potentially diverted
        altitude_min_diverted = np.min(altitude_ft[potentially_diverted])
        mask = within_cruise & is_low_altitude & (altitude_ft == altitude_min_diverted)
        i_lowest_altitude = np.flatnonzero(mask)[0] + 1

        # get reference to "on_ground"
        ground_indicator = messages["on_ground"].to_numpy()

        # Check for flight diversion
        condition_1 = np.any(altitude_ft[within_cruise] < flight.MAX_AIRPORT_ELEVATION)
        condition_2 = np.all(segment_duration[within_cruise] < (15.0 * 60.0))
        condition_3 = np.all(segment_length[within_cruise] > 500.0)
        condition_4 = np.nansum(segment_duration[i_lowest_altitude:]) < (2.0 * 60.0 * 60.0)
        condition_5 = np.all(~ground_indicator[within_cruise])

        flight_diversion = condition_1 & condition_2 & condition_3 & condition_4 & condition_5

        # if there is a potential flight diversion, just return a single flight index
        if flight_diversion:
            return flight_id

    # If there are multiple cruise phases, get cut-off point
    i_cutoff = int(np.argmax(segment_duration[anomalous_phase]) + 1)

    # assign a new id after cutoff
    flight_id.iloc[i_cutoff:] = generate_flight_id(
        messages["timestamp"].iloc[i_cutoff], messages["callsign"].iloc[i_cutoff]
    )

    return flight_id


def validate_flights(messages: pd.DataFrame) -> pd.Series:
    """Validate unique flights from Spire ADS-B messages.

    Parameters
    ----------
    messages : pd.DataFrame
        Messages that has been assigned flight ids by :func:`identify_flights`.
        Requires "flight_id" column to be defined in messages

    Returns
    -------
    pd.Series
        Boolean array of `flight_id` validity with the same index as `messages`

    Notes
    -----
    See :func:`is_valid_trajectory` docstring for validation criteria

    See Also
    --------
    :func:`is_valid_trajectory`
    :func:`identify_flights`
    """
    if "flight_id" not in messages:
        raise KeyError("'flight_id' column required in messages")

    # Set default flight id
    valid = pd.Series(
        data=False,
        dtype=bool,
        index=messages.index,
    )

    for _, gp in messages[
        [
            "flight_id",
            "aircraft_type_icao",
            "timestamp",
            "altitude_baro",
            "on_ground",
            "speed",
        ]
    ].groupby("flight_id", sort=False):
        # save flight ids
        valid.loc[gp.index] = is_valid_trajectory(gp)

    return valid


def is_valid_trajectory(
    messages: pd.DataFrame,
    *,
    minimum_messages: int = TRAJECTORY_MINIMUM_MESSAGES,
    final_time_available: pd.Timestamp | None = None,
) -> bool:
    """
    Ensure messages likely contain only one unique flight trajectory.

    Parameters
    ----------
    messages: pd.DataFrame
        ADS-B messages from a single flight trajectory.
    minimum_messages: int, optional
        Minimum number of messages required for trajectory to be accepted.
        Defaults to :attr:`TRAJECTORY_MINIMUM_MESSAGES`
    final_time_available: pd.Timestamp, optional
        Time of the final recorded ADS-B message available.
        Relaxes the criteria for flight completion (see *Notes*).

    Returns
    -------
    bool
        Boolean indicating if messages constitute a single valid trajectory

    Notes
    -----
    The inputs messages must satisfy all
    the following conditions to validate the flight trajectory:

    #. Must contain more than :attr:`TRAJECTORY_MINIMUM_MESSAGES`
    #. Trajectory must contain a `cruise` phase as defined by :func:`flight.segment_phase`
    #. Trajectory must not go below the minimum cruising altitude throughout the cruise phase.
    #. A trajectory must be "complete", defined when one of these conditions are satisfied:
        - Final message is on the ground, altitude < :attr:`flight.MAX_AIRPORT_ELEVATION` feet
          and speed < :attr:`flight.MAX_ON_GROUND_SPEED` knots
        - Final message is < :attr:`flight.MAX_AIRPORT_ELEVATION` feet, in descent,
          and > 2 h have passed since `final_time_available`.
        - At least 12 h have passed since `final_time_available`
          (remaining trajectory might not be recorded).

    Trajectory is not valid if any criteria is False.
    """

    # Use BADA 3 to calculate min cruise altitude
    try:
        from pycontrails.ext.bada import BADA3

        bada3 = BADA3()
        aircraft_type_icao = (
            messages["aircraft_type_icao"].iloc[0]
            if len(messages["aircraft_type_icao"]) > 1
            else messages["aircraft_type_icao"]
        )

        # Remove erroneous altitude, i.e., altitude above operating limit of aircraft type
        altitude_ceiling_ft = bada3.ptf_param_dict[aircraft_type_icao].max_altitude_ft
        min_cruise_altitude_ft = 0.5 * altitude_ceiling_ft

    except (ImportError, FileNotFoundError, KeyError):
        min_cruise_altitude_ft = flight.MIN_CRUISE_ALTITUDE

    # Flight duration
    segment_duration = flight.segment_duration(messages["timestamp"].to_numpy())
    is_short_haul = np.nansum(segment_duration).item() < flight.SHORT_HAUL_DURATION

    # Flight phase
    altitude_ft = messages["altitude_baro"].to_numpy()
    segment_rocd = flight.segment_rocd(segment_duration, altitude_ft)
    segment_phase = flight.segment_phase(
        segment_rocd,
        altitude_ft,
        threshold_rocd=250.0,
        min_cruise_altitude_ft=min_cruise_altitude_ft,
    )

    # Find any anomalous messages with low altitudes between
    # the start and end of the cruise phase
    # See `_separate_by_cruise_phase` for more comments on logic
    cruise = segment_phase == flight.FlightPhase.CRUISE
    within_cruise = np.bitwise_xor.accumulate(cruise) | cruise
    is_low_altitude = altitude_ft < min_cruise_altitude_ft
    anomalous_phase = within_cruise & is_low_altitude

    # Validate flight trajectory
    has_enough_messages = len(messages) > minimum_messages
    has_cruise_phase = np.any(segment_phase == flight.FlightPhase.CRUISE).item()
    has_no_anomalous_phase = np.any(anomalous_phase).item()

    # Relax constraint for short-haul flights
    if is_short_haul and (not has_cruise_phase) and (not has_no_anomalous_phase):
        has_cruise_phase = np.any(segment_phase == flight.FlightPhase.LEVEL_FLIGHT).item()
        has_no_anomalous_phase = True

    if not (has_enough_messages and has_cruise_phase and has_no_anomalous_phase):
        return False

    # Check that flight is complete
    # First option is the flight is on the ground and low
    final_message = messages.iloc[-1]
    complete_1 = (
        final_message["on_ground"]
        & (final_message["altitude_baro"] < flight.MAX_AIRPORT_ELEVATION)
        & (final_message["speed"] < flight.MAX_ON_GROUND_SPEED)
    )

    # Second option is the flight is in descent and 2 hours of data are available after
    if final_time_available:
        is_descent = np.any(segment_phase[-5:] == flight.FlightPhase.DESCENT) | np.any(
            segment_phase[-5:] == flight.FlightPhase.LEVEL_FLIGHT
        )
        elapsed_time_hrs = (final_time_available - final_message["timestamp"]) / np.timedelta64(
            1, "h"
        )
        complete_2 = (
            (final_message["altitude_baro"] < flight.MAX_AIRPORT_ELEVATION)
            & is_descent
            & (elapsed_time_hrs > 2.0)
        )

        # Third option is 12 hours of data are available after
        complete_3 = elapsed_time_hrs > 12.0
    else:
        complete_2 = False
        complete_3 = False

    # Complete is defined as one of these criteria being satisfied
    is_complete = complete_1 | complete_2 | complete_3

    return has_enough_messages and has_cruise_phase and has_no_anomalous_phase and is_complete


def _downsample_flight(
    messages: pd.DataFrame,
    *,
    time_resolution: str | pd.DateOffset | pd.Timedelta = "10s",
) -> pd.DataFrame:
    """
    Downsample ADS-B messages to a specified time resolution.

    .. warning::
        This function is not used.
        Use :meth:`flight.Flight.resample_and_fill` after creating `Flight`
        instead.

    .. note::
        This function does not interpolate when upsampling.
        Nan values will be dropped.

    Parameters
    ----------
    messages: pd.DataFrame
        ADS-B messages from a single flight trajectory.
    time_resolution: str | pd.DateOffset | pd.Timedelta
        Downsampled time resolution.
        Any input compatible with :meth:`pandas.DataFrame.resample`.
        Defaults to "10s" (10 seconds).

    Returns
    -------
    pd.DataFrame
        Downsampled ADS-B messages for a single flight trajectory.

    See Also
    --------
    :meth:`pandas.DataFrame.resample`
    :meth:`flight.Flight.resample_and_fill`
    """
    mdf = messages.copy()
    mdf.set_index("timestamp", inplace=True, drop=False)
    resampled = mdf.resample(time_resolution).first()

    # remove rows that do not align with a previous time
    resampled = resampled.loc[resampled["longitude"].notna()]

    # reset original index and return
    return resampled.reset_index(drop=True)



"""Tools for spherical geometry, solar radiation, and wind advection."""

from __future__ import annotations

import numpy as np
import xarray as xr

from pycontrails.physics import constants, units
from pycontrails.utils.types import ArrayLike

# ------------------
# Spherical Geometry
# ------------------


def haversine(lons0: ArrayLike, lats0: ArrayLike, lons1: ArrayLike, lats1: ArrayLike) -> ArrayLike:
    r"""Calculate haversine distance between points in (lons0, lats0) and (lons1, lats1).

    Handles coordinates crossing the antimeridian line (-180, 180).

    Parameters
    ----------
    lons0, lats0 : ArrayLike
        Coordinates of initial points, [:math:`\deg`]
    lons1, lats1 : ArrayLike
        Coordinates of terminal points, [:math:`\deg`]

    Returns
    -------
    ArrayLike
        Distances between corresponding points. [:math:`m`]

    Notes
    -----
    This formula does not take into account the non-spheroidal (ellipsoidal) shape of the Earth.
    Originally referenced from https://andrew.hedges.name/experiments/haversine/.

    References
    ----------
    - :cite:`CalculateDistanceBearing`

    See Also
    --------
    :func:`sklearn.metrics.pairwise.haversine_distances`:
        Compute the Haversine distance
    :class:`pyproj.Geod`:
        Performs forward and inverse geodetic, or Great Circle, computations
    """
    lats0_rad = units.degrees_to_radians(lats0)
    lats1_rad = units.degrees_to_radians(lats1)

    cos_lats0 = np.cos(lats0_rad)
    cos_lats1 = np.cos(lats1_rad)

    d_lons = units.degrees_to_radians(lons1) - units.degrees_to_radians(lons0)
    d_lats = lats1_rad - lats0_rad

    a = (np.sin(d_lats / 2.0)) ** 2 + cos_lats0 * cos_lats1 * ((np.sin(d_lons / 2.0)) ** 2)
    cc = 2.0 * np.arctan2(a**0.5, (1.0 - a) ** 0.5)
    return constants.radius_earth * cc


def segment_haversine(longitude: np.ndarray, latitude: np.ndarray) -> np.ndarray:
    r"""Calculate haversine distance between consecutive points along path.

    Parameters
    ----------
    longitude : np.ndarray
        1D Longitude values with index corresponding to latitude inputs, [:math:`\deg`]
    latitude : np.ndarray
        1D Latitude values with index corresponding to longitude inputs, [:math:`\deg`]

    Returns
    -------
    np.ndarray
        Haversine distance between (lat_i, lon_i) and (lat_i+1, lon_i+1), [:math:`m`]
        The final entry of the output is set to nan.

    See Also
    --------
    :meth:`pyproj.Geod.line_lengths`
    """
    dtype = np.result_type(longitude, latitude, np.float32)
    dist = np.empty(longitude.size, dtype=dtype)

    lons0 = longitude[:-1]
    lons1 = longitude[1:]
    lats0 = latitude[:-1]
    lats1 = latitude[1:]

    dist[:-1] = haversine(lons0, lats0, lons1, lats1)
    dist[-1] = np.nan
    return dist


def azimuth_to_direction(
    azimuth_: np.ndarray, latitude: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    r"""Calculate rectangular direction from spherical azimuth.

    This implementation uses the equation

    ``cos(latitude) / tan(azimuth) = sin_a / cos_a``

    to solve for `sin_a` and `cos_a`.

    Parameters
    ----------
    azimuth_ : np.ndarray
        Angle measured clockwise from true north, [:math:`\deg`]
    latitude : np.ndarray
        Latitude value of the point, [:math:`\deg`]

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        A tuple of sine and cosine values.
    """
    cos_lat = np.cos(units.degrees_to_radians(latitude))
    tan_az = np.tan(units.degrees_to_radians(azimuth_))

    num = cos_lat
    denom = tan_az
    mag = np.sqrt(num**2 + denom**2)

    # For azimuth in [0, 90) and (270, 360], sin_a positive
    sign_sin_a = np.where((azimuth_ - 90.0) % 360.0 - 180.0 >= 0.0, 1.0, -1.0)

    # For azimuth in [0, 180), cos_a positive
    sign_cos_a = np.where(azimuth_ % 360.0 - 180.0 <= 0.0, 1.0, -1.0)

    sin_a = sign_sin_a * np.abs(num) / mag
    cos_a = sign_cos_a * np.abs(denom) / mag
    return sin_a, cos_a


def azimuth(
    lons0: np.ndarray,
    lats0: np.ndarray,
    lons1: np.ndarray,
    lats1: np.ndarray,
) -> np.ndarray:
    r"""Calculate angle relative to true north for set of coordinates.

    Parameters
    ----------
    lons0 : np.ndarray
        Longitude values of initial endpoints, [:math:`\deg`].
    lats0 : np.ndarray
        Latitude values of initial endpoints, [:math:`\deg`].
    lons1 : np.ndarray
        Longitude values of terminal endpoints, [:math:`\deg`].
    lats1 : np.ndarray
        Latitude values of terminal endpoints, [:math:`\deg`].

    References
    ----------
    - :cite:`wikipediacontributorsAzimuth2023`

    Returns
    -------
    np.ndarray
        Azimuth relative to true north (:math:`0\deg`), [:math:`\deg`]

    See Also
    --------
    :func:`longitudinal_angle`
    """
    lons0 = units.degrees_to_radians(lons0)
    lons1 = units.degrees_to_radians(lons1)
    lats0 = units.degrees_to_radians(lats0)
    lats1 = units.degrees_to_radians(lats1)
    d_lon = lons1 - lons0

    num = np.sin(d_lon)
    denom = np.cos(lats0) * np.tan(lats1) - np.sin(lats0) * np.cos(d_lon)

    # outputs on [-180, 180] range
    alpha = units.radians_to_degrees(np.arctan2(num, denom))

    # return on [0, 360)
    return alpha % 360.0


def segment_azimuth(longitude: np.ndarray, latitude: np.ndarray) -> np.ndarray:
    r"""Calculate the angle between coordinate segments and true north.

    `np.nan` is added to the final value so the length of the output is the same as the inputs.

    Parameters
    ----------
    longitude : np.ndarray
        Longitude values, [:math:`\deg`]
    latitude : np.ndarray
        Latitude values, [:math:`\deg`]

    Returns
    -------
    np.ndarray
        Azimuth relative to true north (:math:`0\deg`), [:math:`\deg`]
        Final entry of each array is set to `np.nan`.

    References
    ----------
    - :cite:`wikipediacontributorsAzimuth2023`

    See Also
    --------
    :func:`azimuth`
    """
    dtype = np.result_type(longitude, latitude, np.float32)
    az = np.empty(longitude.size, dtype=dtype)

    lons0 = longitude[:-1]
    lons1 = longitude[1:]
    lats0 = latitude[:-1]
    lats1 = latitude[1:]

    az[:-1] = azimuth(lons0, lats0, lons1, lats1)
    az[-1] = np.nan
    return az


def longitudinal_angle(
    lons0: np.ndarray,
    lats0: np.ndarray,
    lons1: np.ndarray,
    lats1: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    r"""Calculate angle with longitudinal axis for sequence of segments.

    Parameters
    ----------
    lons0 : np.ndarray
        Longitude values of initial endpoints, [:math:`\deg`].
    lats0 : np.ndarray
        Latitude values of initial endpoints, [:math:`\deg`].
    lons1 : np.ndarray
        Longitude values of terminal endpoints, [:math:`\deg`].
    lats1 : np.ndarray
        Latitude values of terminal endpoints, [:math:`\deg`].

    References
    ----------
    - :cite:`wikipediacontributorsAzimuth2023`

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Sine, cosine values.
    """
    lons0 = units.degrees_to_radians(lons0)
    lons1 = units.degrees_to_radians(lons1)
    lats0 = units.degrees_to_radians(lats0)
    lats1 = units.degrees_to_radians(lats1)
    d_lon = lons1 - lons0

    num = np.sin(d_lon)
    denom = np.cos(lats0) * np.tan(lats1) - np.sin(lats0) * np.cos(d_lon)
    mag = np.sqrt(num**2 + denom**2)

    where = mag > 0.0
    out = np.full_like(mag, np.nan)

    sin_a = np.divide(denom, mag, out=out.copy(), where=where)
    cos_a = np.divide(num, mag, out=out.copy(), where=where)
    return sin_a, cos_a


def segment_angle(longitude: np.ndarray, latitude: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    r"""Calculate the angle between coordinate segments and the longitudinal axis.

    `np.nan` is added to the final value so the length of the output is the same as the inputs.

    Parameters
    ----------
    longitude : np.ndarray
        Longitude values, [:math:`\deg`]
    latitude : np.ndarray
        Latitude values, [:math:`\deg`]

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        sin(a), cos(a), where ``a`` is the angle between the segment and the longitudinal axis.
        Final entry of each array is set to `np.nan`.

    References
    ----------
    - :cite:`wikipediacontributorsAzimuth2023`

    Notes
    -----
    ::

            (lon_2, lat_2)  X
                           /|
                          / |
                         /  |
                        /   |
                       /    |
                      /     |
                     /      |
    (lon_1, lat_1)  X -------> longitude (x-axis)

    See Also
    --------
    :func:`longitudinal_angle`
    """
    dtype = np.result_type(longitude, latitude, np.float32)
    sin_a = np.empty(longitude.size, dtype=dtype)
    cos_a = np.empty(longitude.size, dtype=dtype)

    lons0 = longitude[:-1]
    lons1 = longitude[1:]
    lats0 = latitude[:-1]
    lats1 = latitude[1:]

    sin_a[:-1], cos_a[:-1] = longitudinal_angle(lons0, lats0, lons1, lats1)
    sin_a[-1] = np.nan
    cos_a[-1] = np.nan
    return sin_a, cos_a


def segment_length(longitude: np.ndarray, latitude: np.ndarray, altitude: np.ndarray) -> np.ndarray:
    r"""Calculate the segment length between coordinates by assuming a great circle distance.

    Requires coordinates to be in EPSG:4326.
    Lengths are calculated using both horizontal and vertical displacement of segments.

    `np.nan` is added to the final value so the length of the output is the same as the inputs.

    Parameters
    ----------
    longitude : np.ndarray
        Longitude values, [:math:`\deg`]
    latitude : np.ndarray
        Latitude values, [:math:`\deg`]
    altitude : np.ndarray
        Altitude values, [:math:`m`]

    Returns
    -------
    np.ndarray
        Array of distances in [:math:`m`] between coordinates.
        Final entry of each array is set to `np.nan`.

    See Also
    --------
    :func:`haversine`
    :func:`segment_haversine`
    """
    dist_horizontal = segment_haversine(longitude, latitude)
    dist_vertical = np.empty_like(altitude)
    dist_vertical[:-1] = np.diff(altitude)
    dist_vertical[-1] = np.nan  # last segment is set to nan
    return (dist_horizontal**2 + dist_vertical**2) ** 0.5


def forward_azimuth(
    lons: np.ndarray, lats: np.ndarray, az: np.ndarray | float, dist: np.ndarray | float
) -> tuple[np.ndarray, np.ndarray]:
    r"""Calculate coordinates along forward azimuth.

    This function is identical to the `pyproj.Geod.fwd` method when working on
    a spherical earth. Both signatures are also identical. This implementation
    is generally more performant.

    Parameters
    ----------
    lons : np.ndarray
        Array of longitude values.
    lats : np.ndarray
        Array of latitude values.
    az : np.ndarray | float
        Azimuth, measured in [:math:`\deg`].
    dist : np.ndarray | float
        Distance [:math:`m`] between initial longitude latitude values and
        point to be computed.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Tuple of longitude latitude arrays.

    See Also
    --------
    :meth:pyproj.Geod.fwd
    """
    az_rad = units.degrees_to_radians(az)
    sin_az = np.sin(az_rad)
    cos_az = np.cos(az_rad)

    lats_rad = units.degrees_to_radians(lats)
    sin_lats = np.sin(lats_rad)
    cos_lats = np.cos(lats_rad)

    dist_ratio = dist / constants.radius_earth
    cos_dist_ratio = np.cos(dist_ratio)
    sin_dist_ratio = np.sin(dist_ratio)

    dest_lats_rad = np.arcsin(sin_lats * cos_dist_ratio + cos_lats * sin_dist_ratio * cos_az)
    dest_lats = units.radians_to_degrees(dest_lats_rad)

    delta_lons_rad = np.arctan2(
        sin_az * sin_dist_ratio * cos_lats,
        cos_dist_ratio - sin_lats * np.sin(dest_lats_rad),
    )
    dest_lons = lons + units.radians_to_degrees(delta_lons_rad)
    dest_lons = (dest_lons + 180.0) % 360.0 - 180.0

    return dest_lons, dest_lats


# ---------------
# Solar Radiation
# ---------------


def solar_direct_radiation(
    longitude: ArrayLike, latitude: ArrayLike, time: ArrayLike, threshold_cos_sza: float = 0.0
) -> np.ndarray:
    r"""Calculate the instantaneous theoretical solar direct radiation (SDR).

    Parameters
    ----------
    longitude : ArrayLike
        Longitude, [:math:`\deg`]
    latitude : ArrayLike
        Latitude, [:math:`\deg`]
    time : ArrayLike
        Time, formatted as :class:`np.datetime64`
    threshold_cos_sza : float, optional
        Set the SDR to 0 when the :func:`cosine_solar_zenith_angle` is below a certain value.
        By default, set to 0.

    Returns
    -------
    ArrayLike
        Solar direct radiation of incoming radiation, [:math:`W m^{-2}`]

    References
    ----------
    - :cite:`UOSRMLSolar`
    """
    theta_rad = orbital_position(time)

    # Use longitude and latitude to determine the dtype
    dtype = np.result_type(longitude, latitude)
    theta_rad = theta_rad.astype(dtype, copy=False)

    _solar_constant = solar_constant(theta_rad)
    cos_sza = cosine_solar_zenith_angle(longitude, latitude, time, theta_rad)

    # Note that np.where is more performant than xr.where, even for large arrays
    # (and especially for small arrays).
    # BUT xr.where is "safer" in the sense that it will pass numpy arrays through as if
    # they were pumped directly through np.where.
    # For now, explicitly check if we're work with xarray instances or numpy arrays
    # This will likely not work for native python numeric types
    if isinstance(cos_sza, xr.DataArray):
        return xr.where(cos_sza < threshold_cos_sza, 0.0, cos_sza * _solar_constant)
    return np.where(cos_sza < threshold_cos_sza, 0.0, cos_sza * _solar_constant)


def solar_constant(theta_rad: ArrayLike) -> ArrayLike:
    """Calculate the solar electromagnetic radiation per unit area from orbital position.

    On average, the extraterrestrial irradiance is 1367 W/m**2
    and varies by +- 3% as the Earth orbits the sun.

    Parameters
    ----------
    theta_rad : ArrayLike
        Orbital position, [:math:`rad`]. Use :func:`orbital_position` to calculate
        the orbital position from time input.

    Returns
    -------
    ArrayLike
        Solar constant, [:math:`W m^{-2}`]

    References
    ----------
    - :cite:`UOSRMLSolar`
    - :cite:`paltridgeRadiativeProcessesMeteorology1976`
    - :cite:`duffieSolarEngineeringThermal1991`

    Notes
    -----
    :math:`orbital_effect = (R_{av} / R)^{2}`
    where :math:`R` is the separation of Earth from the sun
    and :math:`R_{av}`is the mean separation.
    """
    orbital_effect = (
        1.00011
        + (0.034221 * np.cos(theta_rad))
        + (0.001280 * np.sin(theta_rad))
        + (0.000719 * np.cos(theta_rad * 2))
        + (0.000077 * np.sin(theta_rad * 2))
    )

    return constants.solar_constant * orbital_effect


def cosine_solar_zenith_angle(
    longitude: ArrayLike,
    latitude: ArrayLike,
    time: ArrayLike,
    theta_rad: ArrayLike,
) -> ArrayLike:
    r"""Calculate the cosine of the solar zenith angle.

    Return (:math:`\cos(\theta)`), where :math:`\theta` is the angle between the sun and the
    vertical direction.

    Parameters
    ----------
    longitude : ArrayLike
        Longitude, [:math:`\deg`]
    latitude : ArrayLike
        Latitude, [:math:`\deg`]
    time : ArrayLike
        Time, formatted as :class:`np.datetime64`
    theta_rad : ArrayLike
        Orbital position, [:math:`rad`]. Output of :func:`orbital_position`.

    Returns
    -------
    ArrayLike
        Cosine of the solar zenith angle

    References
    ----------
    - :cite:`wikipediacontributorsSolarZenithAngle2022`

    See Also
    --------
    :func:`orbital_position`
    :func:`solar_declination_angle`
    :func:`solar_hour_angle`
    """
    lat_rad = units.degrees_to_radians(latitude)
    sdec_rad = units.degrees_to_radians(solar_declination_angle(theta_rad))
    sha_rad = units.degrees_to_radians(solar_hour_angle(longitude, time, theta_rad))

    return np.sin(lat_rad) * np.sin(sdec_rad) + (
        np.cos(lat_rad) * np.cos(sdec_rad) * np.cos(sha_rad)
    )


def orbital_position(time: ArrayLike) -> ArrayLike:
    """Calculate the orbital position of Earth to a reference point set at the start of year.

    Parameters
    ----------
    time : ArrayLike
        ArrayLike of :class:`np.datetime64` times

    Returns
    -------
    ArrayLike
        Orbital position of Earth, [:math:`rad`]
    """
    dt_day = days_since_reference_year(time)
    theta = 360.0 * (dt_day / 365.25)
    return units.degrees_to_radians(theta)


def days_since_reference_year(time: ArrayLike, ref_year: int = 2000) -> ArrayLike:
    """Calculate the days elapsed since the start of the reference year.

    Parameters
    ----------
    time : ArrayLike
        ArrayLike of :class:`np.datetime64` times
    ref_year : int, optional
        Year of reference

    Returns
    -------
    ArrayLike
        Days elapsed since the reference year. Output ``dtype`` is ``np.float64``.

    Raises
    ------
    RuntimeError
        Raises when reference year is greater than the time of `time` element
    """
    date_start = np.datetime64(ref_year - 1970, "Y")
    dt_day = (time - date_start) / np.timedelta64(1, "D")

    if np.any(dt_day < 0.0):
        raise RuntimeError(
            f"Reference year {ref_year} is greater than the time of one or more waypoints."
        )

    return dt_day


def hours_since_start_of_day(time: ArrayLike) -> ArrayLike:
    """Calculate the hours elapsed since the start of day (00:00:00 UTC).

    Parameters
    ----------
    time : ArrayLike
        ArrayLike of :class:`np.datetime64` times

    Returns
    -------
    ArrayLike
        Hours elapsed since the start of today day. Output ``dtype`` is ``np.float64``.
    """
    return (time - time.astype("datetime64[D]")) / np.timedelta64(1, "h")


def solar_declination_angle(theta_rad: ArrayLike) -> ArrayLike:
    r"""Calculate the solar declination angle from the orbital position in radians (theta_rad).

    The solar declination angle is the angle between the rays of the Sun and the plane of the
    Earth's equator.

    It has a range of between -23.5 (winter solstice) and +23.5 (summer solstice) degrees.

    Parameters
    ----------
    theta_rad : ArrayLike
        Orbital position, [:math:`rad`]. Output of :func:`orbital_position`.

    Returns
    -------
    ArrayLike
        Solar declination angle, [:math:`\deg`]

    References
    ----------
    - :cite:`paltridgeRadiativeProcessesMeteorology1976`

    Notes
    -----
    Tested against :cite:`noaaSolarCalculationDetails`

    See Also
    --------
    :func:`orbital_position`
    :func:`cosine_solar_zenith_angle`
    """
    return (
        0.396372
        - (22.91327 * np.cos(theta_rad))
        + (4.02543 * np.sin(theta_rad))
        - (0.387205 * np.cos(2 * theta_rad))
        + (0.051967 * np.sin(2 * theta_rad))
        - (0.154527 * np.cos(3 * theta_rad))
        + (0.084798 * np.sin(3 * theta_rad))
    )


def solar_hour_angle(longitude: ArrayLike, time: ArrayLike, theta_rad: ArrayLike) -> ArrayLike:
    r"""Calculate the sun's East to West angular displacement around the polar axis.

    The solar hour angle is an expression of time in angular measurements:
    the value of the hour angle is zero at noon,
    negative in the morning, and positive in the afternoon, increasing by 15 degrees per hour.

    Parameters
    ----------
    longitude : ArrayLike
        Longitude, [:math:`\deg`]
    time : ArrayLike
        ArrayLike of :class:`np.datetime64` times
    theta_rad : ArrayLike
        Orbital position, [:math:`rad`]. Output of :func:`orbital_position`.

    Returns
    -------
    ArrayLike
        Solar hour angle, [:math:`\deg`]

    See Also
    --------
    :func:`orbital_position`
    :func:`cosine_solar_zenith_angle`
    :func:`orbital_correction_for_solar_hour_angle`
    """
    # Let the two float-like arrays dictate the dtype of the time conversion
    dtype = np.result_type(longitude, theta_rad)
    dt_hour = hours_since_start_of_day(time).astype(dtype)

    orbital_correction = orbital_correction_for_solar_hour_angle(theta_rad)
    return ((dt_hour - 12) * 15) + longitude + orbital_correction


def orbital_correction_for_solar_hour_angle(theta_rad: ArrayLike) -> ArrayLike:
    r"""Calculate correction to the solar hour angle due to Earth's orbital location.

    Parameters
    ----------
    theta_rad : ArrayLike
        Orbital position, [:math:`rad`]

    Returns
    -------
    ArrayLike
        Correction to the solar hour angle as a result of Earth's orbital location, [:math:`\deg`]

    References
    ----------
    - :cite:`paltridgeRadiativeProcessesMeteorology1976`

    Notes
    -----
    Tested against :cite:`noaaSolarCalculationDetails`
    """
    return (
        0.004297
        + (0.107029 * np.cos(theta_rad))
        - (1.837877 * np.sin(theta_rad))
        - (0.837378 * np.cos(2 * theta_rad))
        - (2.340475 * np.sin(2 * theta_rad))
    )


# ---------
# Advection
# ---------


def advect_longitude(
    longitude: ArrayLike, latitude: ArrayLike, u_wind: ArrayLike, dt: np.ndarray | np.timedelta64
) -> ArrayLike:
    r"""Calculate the longitude of a particle after time `dt` caused by advection due to wind.

    Automatically wrap over the antimeridian if necessary.

    Parameters
    ----------
    longitude : ArrayLike
        Original longitude, [:math:`\deg`]
    latitude : ArrayLike
        Original latitude, [:math:`\deg`]
    u_wind : ArrayLike
        Wind speed in the longitudinal direction, [:math:`m s^{-1}`]
    dt : np.ndarray
        Advection timestep

    Returns
    -------
    ArrayLike
        New longitude value, [:math:`\deg`]
    """
    # Use the same dtype as longitude, latitude, and u_wind
    dtype = np.result_type(longitude, latitude, u_wind)
    dt_s = _dt_to_float_seconds(dt, dtype)

    distance_m = u_wind * dt_s

    new_longitude = longitude + units.m_to_longitude_distance(distance_m, latitude)
    return (new_longitude + 180.0) % 360.0 - 180.0  # wrap antimeridian


def advect_latitude(
    latitude: ArrayLike, v_wind: ArrayLike, dt: np.ndarray | np.timedelta64
) -> ArrayLike:
    r"""Calculate the latitude of a particle after time ``dt`` caused by advection due to wind.

    .. note::

        It is possible for advected latitude values to lie outside of the WGS84 domain
        ``[-90, 90]``. In :class:`Cocip` models, latitude values close to the poles
        create an end of life condition, thereby avoiding this issue. In practice,
        such situations are very rare.

        These polar divergence issues could also be addressed by reflecting the
        longitude values 180 degrees via a spherical equivalence such as
        ``(lon, lat) ~ (lon + 180, 180 - lat)``. This approach is not currently taken.

    Parameters
    ----------
    latitude : ArrayLike
        Original latitude, [:math:`\deg`]
    v_wind : ArrayLike
        Wind speed in the latitudinal direction, [:math:`m s^{-1}`]
    dt : np.ndarray
        Advection time delta

    Returns
    -------
    ArrayLike
        New latitude value, [:math:`\deg`]
    """
    # Use the same dtype as latitude and v_wind
    dtype = np.result_type(latitude, v_wind)
    dt_s = _dt_to_float_seconds(dt, dtype)

    distance_m = v_wind * dt_s

    return latitude + units.m_to_latitude_distance(distance_m)


def advect_level(
    level: ArrayLike,
    vertical_velocity: ArrayLike,
    rho_air: ArrayLike,
    terminal_fall_speed: ArrayLike,
    dt: np.ndarray | np.timedelta64,
) -> ArrayLike:
    r"""Calculate the pressure level of a particle after time ``dt``.

    This function calculates the new pressure level of a particle as a result of
    vertical advection caused by the vertical velocity and terminal fall speed.

    Parameters
    ----------
    level : ArrayLike
        Pressure level, [:math:`hPa`]
    vertical_velocity : ArrayLike
        Vertical velocity, [:math:`m s^{-1}`]
    rho_air : ArrayLike
        Air density, [:math:`kg m^{-3}`]
    terminal_fall_speed : ArrayLike
        Terminal fall speed of the particle, [:math:`m s^{-1}`]
    dt : np.ndarray
        Time delta for each waypoint

    Returns
    -------
    ArrayLike
        New pressure level, [:math:`hPa`]
    """
    dt_s = _dt_to_float_seconds(dt, level.dtype)
    velocity = vertical_velocity + rho_air * terminal_fall_speed * constants.g

    return (level * 100.0 + (dt_s * velocity)) / 100.0


def _dt_to_float_seconds(dt: np.ndarray | np.timedelta64, dtype: np.dtype) -> np.ndarray:
    """Convert a time delta to seconds as a float with specified ``dtype`` precision.

    Parameters
    ----------
    dt : np.ndarray
        Time delta for each waypoint
    dtype : np.dtype
        Data type of the output array

    Returns
    -------
    np.ndarray
        Time delta in seconds as a float
    """
    out = np.empty(dt.shape, dtype=dtype)
    np.divide(dt, np.timedelta64(1, "s"), out=out)
    return out



"""Wind shear functions."""

from __future__ import annotations

import numpy as np

from pycontrails.utils.types import ArrayScalarLike


def wind_shear_enhancement_factor(
    contrail_depth: np.ndarray,
    effective_vertical_resolution: float | np.ndarray,
    wind_shear_enhancement_exponent: float | np.ndarray,
) -> np.ndarray:
    r"""Calculate the multiplication factor to enhance the wind shear based on contrail depth.

    This factor accounts for any subgrid-scale that is not captured by the resolution
    of the meteorological datasets.

    Parameters
    ----------
    contrail_depth : np.ndarray
        Contrail depth , [:math:`m`]. Expected to be positive and bounded away from 0.
    effective_vertical_resolution : float | np.ndarray
        Vertical resolution of met data , [:math:`m`]
    wind_shear_enhancement_exponent : float | np.ndarray
        Exponent used in calculation. Expected to be nonnegative.
        Discussed in paragraphs following eq. (39) in Schumann 2012 and referenced as `n`.
        When this parameter is 0, no enhancement occurs.

    Returns
    -------
    np.ndarray
        Wind shear enhancement factor

    Notes
    -----
    Implementation based on eq (39) in :cite:`schumannContrailCirrusPrediction2012`.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    ratio = effective_vertical_resolution / contrail_depth
    return 0.5 * (1 + ratio**wind_shear_enhancement_exponent)


def wind_shear_normal(
    u_wind_top: ArrayScalarLike,
    u_wind_btm: ArrayScalarLike,
    v_wind_top: ArrayScalarLike,
    v_wind_btm: ArrayScalarLike,
    cos_a: ArrayScalarLike,
    sin_a: ArrayScalarLike,
    dz: float,
) -> ArrayScalarLike:
    r"""Calculate the total wind shear normal to an axis.

    The total wind shear is the vertical gradient of the horizontal velocity.

    Parameters
    ----------
    u_wind_top : ArrayScalarLike
        u wind speed in the top layer, [:math:`m \ s^{-1}`]
    u_wind_btm : ArrayScalarLike
        u wind speed in the bottom layer, [:math:`m \ s^{-1}`]
    v_wind_top : ArrayScalarLike
        v wind speed in the top layer, [:math:`m \ s^{-1}`]
    v_wind_btm : ArrayScalarLike
        v wind speed in the bottom layer, [:math:`m \ s^{-1}`]
    cos_a : ArrayScalarLike
        Cosine component of segment
    sin_a : ArrayScalarLike
        Sine component of segment
    dz : float
        Difference in altitude between measurements, [:math:`m`]

    Returns
    -------
    ArrayScalarLike
       Wind shear normal to axis, [:math:`s^{-1}`]
    """
    du_dz = (u_wind_top - u_wind_btm) / dz
    dv_dz = (v_wind_top - v_wind_btm) / dz
    return dv_dz * cos_a - du_dz * sin_a


def wind_shear(
    u_wind_top: ArrayScalarLike,
    u_wind_btm: ArrayScalarLike,
    v_wind_top: ArrayScalarLike,
    v_wind_btm: ArrayScalarLike,
    dz: float,
) -> ArrayScalarLike:
    r"""Calculate the total wind shear.

    The total wind shear is the vertical gradient of the horizontal velocity.

    Parameters
    ----------
    u_wind_top : ArrayScalarLike
        u wind speed in the top layer, [:math:`m \ s^{-1}`]
    u_wind_btm : ArrayScalarLike
        u wind speed in the bottom layer, [:math:`m \ s^{-1}`]
    v_wind_top : ArrayScalarLike
        v wind speed in the top layer, [:math:`m \ s^{-1}`]
    v_wind_btm : ArrayScalarLike
        v wind speed in the bottom layer, [:math:`m \ s^{-1}`]
    dz : float
        Difference in altitude between measurements, [:math:`m`]

    Returns
    -------
    ArrayScalarLike
       Total wind shear, [:math:`s^{-1}`]
    """
    du_dz = (u_wind_top - u_wind_btm) / dz
    dv_dz = (v_wind_top - v_wind_btm) / dz
    return (du_dz**2 + dv_dz**2) ** 0.5



"""Test `Airports`."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pycontrails.core import airports

try:
    import requests

    r = requests.head("https://github.com", timeout=10)
except Exception:
    offline = True
else:
    offline = False


@pytest.fixture
def airport_database() -> pd.DataFrame:
    """Download copy of airport database."""
    return airports.global_airport_database()


@pytest.mark.skipif(offline, reason="offline")
def test_global_airports_database(airport_database: pd.DataFrame) -> None:
    """Ensure that airport database exists."""
    assert "type" in airport_database
    assert "elevation_m" in airport_database


@pytest.mark.skipif(offline, reason="offline")
def test_nearest_airport_identification(airport_database: pd.DataFrame) -> None:
    """Ensure that the nearest airport for a given coordinates is correctly identified."""
    # Somewhere near Singapore Changi Airport (1.36 N, 103.99 E)
    longitude = 103.5
    latitude = 1.5
    altitude = 2950
    airport_icao = airports.find_nearest_airport(airport_database, longitude, latitude, altitude)
    assert airport_icao == "WSSS"

    # Somewhere near London Heathrow Airport (51.47 N, 0.45 W)
    longitude = -1
    latitude = 51.6
    altitude = 2950
    airport_icao = airports.find_nearest_airport(airport_database, longitude, latitude, altitude)
    assert airport_icao == "EGLL"

    # Somewhere near London Gatwick Airport (51.15 N, 0.18 W)
    longitude = -0.18
    latitude = 50.5
    altitude = 2950
    airport_icao = airports.find_nearest_airport(airport_database, longitude, latitude, altitude)
    assert airport_icao == "EGKK"

    # In the middle of the South Atlantic Ocean
    longitude = -30
    latitude = -60
    altitude = 2950
    airport_icao = airports.find_nearest_airport(airport_database, longitude, latitude, altitude)
    assert airport_icao is None


@pytest.mark.skipif(offline, reason="offline")
def test_invalid_altitude(airport_database: pd.DataFrame) -> None:
    """Ensure that function will not execute if provided altitude is too high."""
    longitude = -1
    latitude = 51.6
    altitude = 5000

    with pytest.raises(ValueError):
        airports.find_nearest_airport(airport_database, longitude, latitude, altitude)


def test_distance_to_airports(airport_database: pd.DataFrame) -> None:
    """Ensure distance to airports are correctly identified."""
    apts = airport_database[airport_database["iata_code"].isin(["BOS", "JFK"])]

    # a point in the massachusetts bay close to BOS
    longitude = -70.842554
    latitude = 42.387466
    altitude = 500
    distance = airports.distance_to_airports(apts, longitude, latitude, altitude)

    assert len(distance) == 2
    assert np.round(distance[0]) == 13616.0  # close to BOS
    assert np.round(distance[1]) == 312342.0  # not terribly far from JFK
